/*****************************************************************************
 *
 * Dictionary Huffman Decompression Code
 *
 ***************************************************************************/
/*
 * This program is copyright, (c)2002.
 *
 * You are granted a license to use this program under the terms of the GNU
 * Public License. You may view the license at the following URL:
 *      http://www.gnu.org/copyleft/gpl.html
 * Note that without accepting this license, you have no right to use this
 * code.
 *
 *
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>

#include "pdbshred.h"
#include "dhdecomp.h"

/****************************************************************************
   DH ('Dictionary Huffman') Decompression
   This is the compression method in Mobipocket 4.4
  
   DH uses huffman codes to represent multiple character strings based on the
   statistical information for the entire file. 
   DH includes the necessary decompression tables along side the compressed 
   text.
   All fields are in big-endian (Motorola) order.

   Future revisions of the format can be expected to compress the HUFF and 
   CDIC record formats.
   
   For the initial version, they are:
 
   Record 'HUFF' - Huffman tables
      0    MAGIC               'HUFF'
      4    Header length       24
      8    Cache Offset        24
     12    Base Table offset   1048
     16    ??? (+256)   
     20    ??? (+1024)
       

   Cache table:  Array  
      Must be 1024 octets in length, 4 octets for each possible octet 
      
      This is a look up table for the length and decoding of short codewords.
      If the codeword represented by the 8 bits is unique, then 0x80 will be 
      set, and the low 5 bits are the length in bits of the code.  The high
      three bytes partially represent the final symbol. 

      If bit 7 is clear, then the code is looked up in the base table

   Base Table: 
      ??? 
   
   Record 'CDIC' - Dictionaries (Usually multiple records)
      0    MAGIC               'CDIC'
      4    Header length       16 
      8    Unknown             ?? 
     12    number of code bits (varies)
   
   Index table: (Immediately after the header)
     	Each WORD is an index further into the dictionary pointing to a symbol
   Symbols:
	WORD    Length   (0x8000 is a flag FIXME)
        OCTETS  'Length' Output symbol
       
****************************************************************************/   


#define MAX(a,b) (((a)>(b))?(a):(b))
#define MIN(a,b) (((a)<(b))?(a):(b))

dh_params dh_param;

int dh_decompress( octet * src, int octets, octet * dst, int avail_in) 
{
	quartet  bits, nextbits, b;
	int	 len, nextlen, wanted, copied;
	quartet	 code, cache_val;
	int      dict, offset, symlength;
	octet	 * p;
	int	 err; 
	int      avail = avail_in;
 
	len = nextlen = 0; 
	bits = nextbits = 0;
	do {
		wanted = 32 - len;
		while (wanted) {
			if (nextlen) {
				copied = MIN(wanted, nextlen);
				bits |= nextbits >> len;
				wanted -= copied;	
				len += copied;
				nextlen -= copied;

				/* GCC problem: (x <<= 32) might be ignored. */ 
				if (copied == 32) nextbits = 0;		      
				else nextbits <<= copied;
			}
			while ( ((32 - nextlen) >= 8) && octets) {
				nextbits |= *(src++) << (24 - nextlen);
				nextlen += 8;
				octets--;
			}
                        if (!octets) break;
		};
		if ((len < 8) && (!bits)) break;

		cache_val = DWORD(dh_param.cache + ((bits >> 24)*4));
		wanted = cache_val & 31;

		if (!wanted) {
			fprintf(stderr,
"DH Failed, Corrupted Huffman Table - encountered code length of zero!\n");
			return -1;
		}
		if (cache_val & 0x80) {
			code = (cache_val >> 8) - (bits >> (32 - wanted));
		} else {
			quartet base_val;

			wanted -= 1;
			do { 
				base_val = DWORD(dh_param.base + (wanted*8));
				code = (bits >> (32 - (wanted+1)));
				wanted++;
				if (wanted > 32) { 
					fprintf(stderr,
"DH Failed, code length would be more than 32 bits! \n");
					return -1;
				}
			} while (base_val > code);
			code = DWORD(dh_param.base + 4 + ((wanted - 1) * 8)) - 
				(bits >> (32 - wanted));
		}
		dict = code >> dh_param.code_length;
		if ((dict > DH_NUM_DICTS) || (!dh_param.dicts[dict])) {
			fprintf(stderr,
"DH Failed, invalid dictionary (%d) from code %lx, bits were: %08lx\n", 
			dict, code, bits);
			return -1;
		}
		code &= (( 1 << (dh_param.code_length)) - 1);	
		offset = WORD(dh_param.dicts[dict] + code * 2);

		if (offset > dh_param.dictsize[dict]) {
			fprintf(stderr,
"DH Failed, corrupted -- code %lx accessed dict %d past the size %d. \n",
			code, dict, dh_param.dictsize[dict]);
		}
		symlength = WORD(dh_param.dicts[dict] + offset);
		p =  dh_param.dicts[dict] + offset + 2; 

		if (!(symlength & 0x8000)) {
			err = dh_decompress(p, symlength, dst, avail);
			if (err < 0) return err;
			dst += err;
			avail -= err;
		} else {
			int i;

			symlength &= 0x7fff;
			if (symlength > 127) {
				fprintf(stderr,
"DH Failed, Length > 127 while handling symbol.  Length = %d. \n", symlength);
				return -1;	
			}
			if (symlength > avail) {
				fprintf(stderr,
"DH Failed, Stream overflowed the buffer space! \n"); 
				return -1;
			}
			memcpy(dst, p, symlength);
			dst += symlength;
			avail -= symlength;
		}
		bits <<= wanted;
		len -= wanted;
		if (len < 0) {
			fprintf(stderr,
"DH Failed, Stream used too many bits, len is %d\n", len);
			return -1; 
		}

	} while (octets || len);

	if ((!octets) && (len) && (bits)) {
		fprintf(stderr, "WARNING: LEFT OVER BITS (%d) Value (%lx)\n",
			len, bits);
	}
	return (avail_in - avail);
} 
